__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from samsa.exceptions import NoAvailablePartitionsError
from samsa.partitioners import random_partitioner
from samsa.partitions import PartitionMap
from samsa.consumer import Consumer
from samsa.utils import attribute_repr


logger = logging.getLogger(__name__)


class TopicMap(object):
    """
    Provides a dictionary-like interface to :class:`~samsa.topics.Topic`
    instances within a cluster.

    :param cluster: The cluster this topic mapping is associated with.
    :type cluster: :class:`samsa.cluster.Cluster`
    """
    def __init__(self, cluster):
        self.cluster = cluster

        self.__topics = {}

    def __getitem__(self, key):
        """
        Returns a :class:`samsa.topics.Topic` for the given key.

        This is a proxy to :meth:`~TopicMap.get` for a more dict-like
        interface.
        """
        return self.get(key)

    def get(self, name):
        """
        Returns a :class:`samsa.topics.Topic` for this topic name, creating a
        new topic if one has not already been registered.
        """
        topic = self.__topics.get(name, None)
        if topic is None:
            topic = self.__topics[name] = Topic(self.cluster, name)
            logger.info('Registered new topic: %s', topic)
        return topic


class Topic(object):
    """
    A topic within a Kafka cluster.

    :param cluster: The cluster that this topic is associated with.
    :type cluster: :class:`samsa.cluster.Cluster`
    :param name: The name of this topic.
    :param partitioner: callable that takes two arguments, ``partitions`` and
        ``key`` and returns a single :class:`~samsa.partitions.Partition`
        instance to publish the message to.
    :type partitioner: any callable type
    """
    def __init__(self, cluster, name, partitioner=random_partitioner):
        self.cluster = cluster
        self.name = name
        self.partitions = PartitionMap(self.cluster, self)
        self.partitioner = partitioner

    __repr__ = attribute_repr('name')

    def latest_offsets(self):
        return [(p.broker.id, p.latest_offset())
                for p
                in self.partitions]

    def publish(self, data, key=None):
        """
        Publishes one or more messages to a random partition of this topic.

        :param data: message(s) to be sent to the broker.
        :type data: ``str`` or sequence of ``str``.
        :param key: a key to be used for semantic partitioning
        :type key: implementation-specific
        """
        if len(self.partitions) < 1:
            raise NoAvailablePartitionsError('No partitions are available to '
                'accept a write for this message. (Is your Kafka broker '
                'running?)')
        partition = self.partitioner(self.partitions, key)
        return partition.publish(data)

    def subscribe(self, group):
        """
        Returns a new consumer that can be used for reading from this topic.

        :param group: the name of the consumer group this consumer belongs to
        :type group: ``str``
        :rtype: :class:`samsa.consumer.consumer.Consumer`
        """
        return Consumer(self.cluster, self, group)



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from kazoo.recipe.watchers import DataWatch, ChildrenWatch
from kazoo.exceptions import NoNodeException

from samsa.client import Client
from samsa.exceptions import ImproperlyConfiguredError
from samsa.utils import attribute_repr


logger = logging.getLogger(__name__)


class BrokerMap(object):
    """
    Represents the topology of all brokers within a Kafka cluster.
    """
    def __init__(self, cluster):
        self.cluster = cluster

        # The internal cache of all brokers available within the cluster.
        self.__brokers = {}

        self._node_path = '/brokers/ids'
        try:
            self._broker_watch = ChildrenWatch(
                self.cluster.zookeeper,
                self._node_path, self._configure
            )
        except NoNodeException:
            raise ImproperlyConfiguredError(
                'The path "%s" does not exist in your '
                'ZooKeeper cluster -- is your Kafka cluster running?' %
                self._node_path)

    def _configure(self, broker_ids):
        """
        Configures the broker mapping.
        """
        # TODO: If this fails, would it make more sense to open a watch on the
        # key, and just return that there are no brokers that are alive, to
        # avoid any race conditions between cluster/application startup?
        logger.info('Refreshing broker configuration from %s...',
            self.cluster.zookeeper)

        alive = set()
        for broker_id in map(int, broker_ids):
            if broker_id not in self.__brokers:
                broker = Broker(self.cluster, id_=broker_id)
                logger.info('Adding new broker to %s: %s', self, broker)
                self.__brokers[broker.id] = broker
            alive.add(broker_id)

        dead = set(self.__brokers.keys()) - alive
        for broker_id in dead:
            broker = self.__brokers[broker_id]
            logger.info('Removing dead broker from %s: %s', self, broker)
            broker.is_dead = True
            del self.__brokers[broker.id]

    # TODO: Add all proxies to appropriate `dict` interface.

    def __len__(self):
        """
        Returns the number of all active brokers.
        """
        return len(self.__brokers)

    def __iter__(self):
        """
        Returns an iterator containing all of the broker IDs within the
        cluster.
        """
        return iter(self.__brokers)

    def __getitem__(self, id_):
        """
        Returns a broker by it's broker ID.
        """
        return self.get(id_)

    def get(self, id_):
        """
        Returns a broker by it's broker ID.
        """
        return self.__brokers[id_]

    def keys(self):
        """
        Returns a list of all broker IDs within the cluster.
        """
        return self.__brokers.keys()

    def values(self):
        """
        Returns a list of every :class:`~samsa.brokers.Broker` within the
        cluster.
        """
        return self.__brokers.values()

    def items(self):
        """
        Returns a list of 2-tuples of the format ``(id, broker)``, where
        ``broker`` is a :class:`~samsa.brokers.Broker` instance.
        """
        return self.__brokers.items()


class Broker(object):
    """
    A Kafka broker.

    :param cluster: The cluster this broker is associated with.
    :type cluster: :class:`samsa.cluster.Cluster`
    :param id_: Kafka broker ID
    """
    def __init__(self, cluster, id_):
        self.cluster = cluster
        self.id = int(id_)

        self.__host = None
        self.__port = None

        self.is_dead = False

        self._node_path = '/brokers/ids/%s' % self.id
        self._config_watcher = DataWatch(
            self.cluster.zookeeper,
            self._node_path, self._configure
        )

    __repr__ = attribute_repr('id')

    def _configure(self, data, stat):
        """
        Configures a broker based on it's state in ZooKeeper.
        """
        logger.info('Retrieved broker data for %s...', self)
        if data is None:
            logger.info('Broker data field was empty. Assuming it was dead.')
            self.__host = self.__port = None
        else:
            creator, self.__host, port = data.split(':')
            self.__port = int(port)

    @property
    def host(self):
        """
        The host that the broker is available at.
        """
        return self.__host

    @property
    def port(self):
        """
        The port that the broker is available at.
        """
        return self.__port

    @property
    def client(self):
        """
        The :class:`samsa.client.Client` object for this broker.

        Only one client is created per broker instance.
        """
        try:
            return self.__client
        except AttributeError:
            self.__client = Client(self.host, self.cluster.handler,
                                   port=self.port)
            return self.__client



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import itertools
import random
import socket
import time

from kazoo.recipe.watchers import ChildrenWatch
from kazoo.exceptions import NoNodeException
from uuid import uuid4

from samsa.config import ConsumerConfig
from samsa.consumer.partitions import PartitionOwnerRegistry
from samsa.exceptions import (SamsaException, NoAvailablePartitionsError,
                              PartitionOwnedError, ImproperlyConfiguredError)

logger = logging.getLogger(__name__)


class Consumer(object):
    """Primary API for consuming kazoo messages as a group.
    """

    def __init__(self, cluster, topic, group):
        """
        :param cluster:
        :type cluster: :class:`samsa.cluster.Cluster`.
        :param topic: The topic to consume messages from.
        :type topic: :class:`samsa.topics.Topic`.
        :param group: The consumer group to join.
        :type group: str.

        """
        self.config = ConsumerConfig().build()
        self.cluster = cluster
        self.topic = topic
        self.group = group
        self.id = "%s:%s" % (socket.gethostname(), uuid4())

        self.id_path = '/consumers/%s/ids' % self.group

        self.partition_owner_registry = PartitionOwnerRegistry(
            self, cluster, topic, group)
        self.partitions = self.partition_owner_registry.get()

        # Keep track of the partition being read and how much has been read
        self._current_partition = None
        self._current_read_ct = 0

        # Watches
        # TODO: This is a *ton* of watches, some of which are duplicated
        #       elsewhere. This should be cleaned up and all watches put
        #       in a single zookeeper connector, like in the Scala driver.
        self._broker_watcher = None
        self._consumer_watcher = None
        self._topic_watcher = None
        self._topics_watcher = None
        self._rebalancing = True # To stop rebalance while setting watches

        self._add_self()


    def _add_self(self):
        """Add this consumer to the zookeeper participants.

        Ensures we don't add more participants than partitions
        """
        for i in xrange(self.config['consumer_retries_max'] or 1):
            time.sleep(i**2) # first run is 0, ensures we sleep before retry

            participants = self._get_participants()
            if len(self.topic.partitions) > len(participants):
                break # some room to spare
            else:
                logger.debug("More consumers than partitions. "
                             "Waiting %is to retry" % (i+1) ** 2)
        else:
            raise NoAvailablePartitionsError("Couldn't acquire partition. "
                                             "More consumers than partitions.")

        path = '%s/%s' % (self.id_path, self.id)
        self.cluster.zookeeper.create(
            path, self.topic.name, ephemeral=True, makepath=True)

        # Set all our watches and then rebalance
        self._rebalancing = False
        broker_path = '/brokers/ids'
        try:
            self._broker_watcher = ChildrenWatch(
                self.cluster.zookeeper, broker_path,
                self._brokers_changed
            )
        except NoNodeException:
            raise ImproperlyConfiguredError(
                'The broker_path "%s" does not exist in your '
                'ZooKeeper cluster -- is your Kafka cluster running?'
                % broker_path)

        topics_path = '/brokers/topics'
        self._topics_watcher = ChildrenWatch(
            self.cluster.zookeeper,
            '/brokers/topics',
            self._topics_changed
        )
        self._rebalancing = True

        # Final watch will trigger rebalance
        self._consumer_watcher = ChildrenWatch(
            self.cluster.zookeeper, self.id_path,
            self._consumers_changed
        )


    def _brokers_changed(self, brokers):
        """Watcher for consumer group changes"""
        if not self._rebalancing:
            return
        logger.info("Rebalance triggered by /brokers/ids change")
        self._rebalance(self.cluster.zookeeper.get_children(self.id_path))


    def _consumers_changed(self, consumer_ids):
        """Watcher for consumer group changes

        """
        if not self._rebalancing:
            return
        logger.info("Rebalance triggered by %s change" % self.id_path)
        self._rebalance(consumer_ids)


    def _topic_changed(self, broker_ids):
        """Watcher for brokers/partition count for a topic

        """
        if not self._rebalancing:
            return
        topic_path = '/brokers/topics/%s' % self.topic.name
        logger.info("Rebalance triggered by %s change" % topic_path)
        self._rebalance(self.cluster.zookeeper.get_children(self.id_path))

    def _topics_changed(self, topics):
        """Watch for the topic we want to show up, then stop watch

        """
        if self.topic.name in topics:
            self._topic_watcher = ChildrenWatch(
                self.cluster.zookeeper,
                '/brokers/topics/%s' % self.topic.name,
                self._topic_changed
            )
            return False # stop watch

    def _get_participants(self, consumer_ids=None):
        """Get a the other consumers of this topic

        :param consumer_ids: List of consumer_ids (from ChildrenWatch)
        """
        zk = self.cluster.zookeeper
        if not consumer_ids:
            try:
                consumer_ids = zk.get_children(self.id_path)
            except NoNodeException:
                logger.debug("Consumer group doesn't exist. "
                             "No participants to find")
                return []

        participants = []
        for id_ in consumer_ids:
            try:
                topic, stat = zk.get("%s/%s" % (self.id_path, id_))
                if topic == self.topic.name:
                    participants.append(id_)
            except NoNodeException:
                pass # disappeared between ``get_children`` and ``get``
        participants.sort()
        return participants


    def _decide_partitions(self, participants):
        """Use consumers and partitions to determined owned partitions

        Give a set of subscribed consumers, every individual consumer should
        be able to figure out the same distribution of available partitions.

        It's very, very important this gives the same result on all machines,
        so things like participant and partition lists are always sorted.

        The algorithm is to distribute blocks of partitions based on
        how many participants there are. If there are partitions remaining,
        the last R participants get one extra, where R is the remainder.
        """
        # Freeze and sort partitions so we always have the same results
        p_to_str = lambda p: '-'.join(
            [p.topic.name, str(p.broker.id), str(p.number)]
        )
        all_partitions = list(self.topic.partitions)
        all_partitions.sort(key=p_to_str)

        # get start point, # of partitions, and remainder
        idx = participants.index(self.id)
        parts_per_consumer = len(all_partitions) / len(participants)
        remainder_ppc = len(all_partitions) % len(participants)

        start = parts_per_consumer * idx + min(idx, remainder_ppc)
        num_parts = parts_per_consumer + (0 if (idx + 1 > remainder_ppc) else 1)

        # assign partitions from i*N to (i+1)*N - 1 to consumer Ci
        new_partitions = itertools.islice(
            all_partitions,
            start,
            start + num_parts
        )
        new_partitions = set(new_partitions)
        logger.info(
            'Rebalancing %i participants for %i partitions. '
            'My Partitions: %s -- Consumers: %s --- All Partitions: %s',
            len(participants), len(all_partitions),
            [p_to_str(p) for p in new_partitions],
            str(participants),
            [p_to_str(p) for p in all_partitions]
        )
        return new_partitions


    def _rebalance(self, consumer_ids):
        """Joins a consumer group and claims partitions.

        """
        logger.info('Rebalancing consumer %s for topic %s.' % (
            self.id, self.topic.name)
        )

        participants = self._get_participants(consumer_ids=consumer_ids)
        new_partitions = self._decide_partitions(participants)

        for i in xrange(self.config['rebalance_retries_max'] or 1):
            if i > 0:
                logger.debug("Retrying in %is" % ((i+1) ** 2))
                time.sleep(i ** 2)
                # Make sure nothing's changed while we waited
                participants = self._get_participants()
                new_partitions = self._decide_partitions(participants)

            # Remove old partitions and acquire new ones.
            old_partitions = self.partitions - new_partitions
            self.stop_partitions(partitions=old_partitions)
            self.partition_owner_registry.remove(old_partitions)

            try:
                self.partition_owner_registry.add(
                    new_partitions - self.partitions
                )
                break
            except PartitionOwnedError, e:
                logger.debug("Someone still owns partition %s.", e)
                continue
        else:
            raise SamsaException("Couldn't acquire partitions.")


    def __iter__(self):
        """Iterator for next_message. Blocks until a message arrives

        """
        while True:
            yield self.next_message()


    def next_message(self, block=True, timeout=None):
        """Get the next message from one of the partitions.

        This works by picking a partition, reading a bunch of messages
        from it, and then switching to the next partition with messages
        ready. To configure how many are read at a time, use the
        setting reads_per_partition.

        Returns None if no messages are ready and timeout expires

        :param block: Whether to block while waiting for a message
        :param timeout: How long to wait, in seconds, if blocking

        """
        while True:
            if not self._current_partition or self._current_partition.empty():
                # Nothing else blocks, so only this needs a timeout
                self._current_partition = self.partition_owner_registry.get_ready_partition(
                    block=block, timeout=timeout
                )
                if not self._current_partition:
                    return None # Nothing ready to read

            message = self._current_partition.next_message(block=False)
            self._current_read_ct += 1
            rpp = self.config['reads_per_partition']

            if message is None or self._current_read_ct > rpp:
                # Queue is either empty or its time to switch
                self._current_partition = None
                self._current_read_ct = 0
                if message is None:
                    continue # empty -- need a new partition

            return message


    def commit_offsets(self):
        """Commit the offsets of all messages consumed so far.

        """
        partitions = list(self.partitions) # freeze in case of rebalance
        for partition in partitions:
            partition.commit_offset()


    def stop_partitions(self, partitions=None):
        """Stop partitions from fetching more threads.

        :param partitions: Partitions to remove. (default: self.partitions)

        """
        if partitions is None:
            partitions = list(self.partitions) # freeze in case of rebalance
        for partition in partitions:
            partition.stop()
            partition.commit_offset()


    def empty(self):
        return all([p.empty() for p in self.partitions])



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import mock
import unittest2

from samsa import brokers
from samsa.cluster import Cluster
from samsa.exceptions import NoAvailablePartitionsError
from samsa.test.integration import FasterKafkaIntegrationTestCase
from samsa.topics import TopicMap, Topic


class TopicIntgrationTestCase(FasterKafkaIntegrationTestCase):
    def test_no_partitions(self):
        # Kill the broker and try to publish something
        topic = self.kafka_cluster.topics['topic']
        self.kafka_broker.stop()
        with self.assertRaises(NoAvailablePartitionsError):
            topic.publish('message')

        # Restart it afterwards, we're on a shared broker
        self.kafka_broker = self.start_broker(self.client, self.hosts)

    def test_get_latest_offsets(self):
        topic = self.get_topic()
        self.assertEqual(topic.latest_offsets(), [(0, 0)])
        topic.publish(['hello!'])
        self.assertEqual(topic.latest_offsets(), [(0, 15)])


class TopicMapTest(unittest2.TestCase):
    def test_get_topic(self):
        topics = TopicMap(cluster=mock.Mock())
        with mock.patch('samsa.partitions.DataWatch'):
            topic = topics.get('topic-1')
        self.assertIsInstance(topic, Topic)

        # Retrieving the topic again should return the same object instance.
        self.assertIs(topic, topics.get('topic-1'))


class PartitionMapTest(unittest2.TestCase):

    def setUp(self):
        self.cluster = mock.Mock(spec=Cluster)
        self.cluster.zookeeper = mock.Mock()

    @mock.patch('samsa.brokers.ChildrenWatch')
    def test_configuration_no_node(self, cw):
        """Test that we get al the brokers. TODO"""

        broker_map = brokers.BrokerMap(self.cluster)
        self.cluster.brokers = broker_map
        with mock.patch('samsa.brokers.DataWatch'):
            broker_map._configure(['0', '1', '2'])
        with mock.patch('samsa.partitions.DataWatch'):
            topic = Topic(self.cluster, 'topic')

        self.assertEqual(len(topic.partitions), len(broker_map))

    @mock.patch('samsa.brokers.ChildrenWatch')
    def test_configuration_with_nodes(self, cw):
        nodes = {
            '0': '5',
            '1': '2',
            '3': '4',
        }

        def get_node_data(path):
            id = path.rsplit('/', 1)[-1]
            return nodes[str(id)]

        broker_map = brokers.BrokerMap(self.cluster)
        self.cluster.brokers = broker_map
        with mock.patch('samsa.brokers.DataWatch'):
            # Tell the BrokerMap which brokers it's managing.
            broker_map._configure(nodes.keys())

        with mock.patch('samsa.partitions.DataWatch'):
            topic = Topic(self.cluster, 'topic')
            # Tell the PartitionMap which brokers it knows of.
            topic.partitions._configure(nodes.keys())
            # Tell each PartitionSet how many partitions it's managing.
            for partition in topic.partitions._partition_sets:
                partition._configure(
                    nodes[str(partition.broker.id)],
                    mock.Mock()
                )

        self.assertEqual(len(topic.partitions), sum(map(int, nodes.values())))



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from samsa.exceptions import SocketDisconnectedError


def recvall_into(socket, bytea):
    """
    Reads enough data from the socket to fill the provided bytearray (modifies
    in-place.)

    This is basically a hack around the fact that ``socket.recv_into`` doesn't
    allow buffer offsets.

    :type socket: :class:`socket.Socket`
    :type bytea: ``bytearray``
    :rtype: ``bytearray``
    """
    offset = 0
    size = len(bytea)
    while offset < size:
        remaining = size - offset
        chunk = socket.recv(remaining)
        if not len(chunk):
            raise SocketDisconnectedError
        bytea[offset:(offset + len(chunk))] = chunk
        offset += len(chunk)
    return bytea


def recv_struct(socket, struct):
    """
    Reads enough data from the socket to unpack the given struct.

    :type socket: :class:`socket.Socket`
    :type struct: :class:`struct.Struct`
    :rtype: ``tuple``
    """
    bytea = bytearray(struct.size)
    recvall_into(socket, bytea)
    return struct.unpack_from(buffer(bytea))


def recv_framed(socket, framestruct):
    """
    :type socket: :class:`socket.Socket`
    :type frame: :class:`struct.Struct`
    :rtype: ``bytearray``
    """
    (size,) = recv_struct(socket, framestruct)
    return recvall_into(socket, bytearray(size))



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# TODO: Replace all this by using class ctors and method arguments with
# appropriate defaults. Document in the code.


class Config(object):

    @classmethod
    def build(cls, kwargs=None, validate=True):
        """Update cls attrs with kwargs and return the resulting dict.
        """

        config = {}

        for i in cls.__dict__:
            if not i.startswith('__'):
                config[i] = getattr(cls, i)

        for k in kwargs or []:
            if validate and k not in cls.__dict__:
                raise AttributeError("%s not in %s." % (k, cls.__name__))
            config[k] = kwargs[k]

        return config


class ConsumerConfig(Config):
    # All times in seconds

    # controls the socket timeout for network requests
    socket_timeout = 30

    # controls the socket receive buffer for network requests
    socket_buffersize = 64 * 1024

    # controls the number of bytes of messages to attempt to fetch in one
    # request to the Kafka server
    fetch_size = 300 * 1024

    # This parameter avoids repeatedly polling a broker node which has no new
    # data. We will backoff every time we get an empty set from the broker for
    # this time period
    # TODO: implement this.
    backoff_increment = 1

    # the high level consumer buffers the messages fetched from the server
    # internally in blocking queues. This parameter controls the size of those
    # queues
    queuedchunks_max = 100

    # if set to true, the consumer periodically commits to zookeeper the latest
    # consumed offset of each partition.
    autocommit_enable = True

    # is the frequency that the consumed offsets are committed to zookeeper.
    autocommit_interval = 10

    # smallest: automatically reset the offset to the smallest offset available
    # on the broker.
    # largest: automatically reset the offset to the largest offset available
    # on the broker.
    # anything else: throw an exception to the consumer.
    autooffset_reset = 'smallest'

    # max number of retries during rebalance
    rebalance_retries_max = 4

    # max number of times to connect as consumer
    # This exists because a quick restart of a consumer will try to connect
    # before the dead one times out. The default of 5 ensures we can outwait
    # that and connect asap.
    consumer_retries_max = 4

    # Number of messages to read from a single partition before
    # moving to the next one with waiting messages. Setting to 1 will
    # round-robin partitions, but can be slow. Default is the same as
    # queuedchunks_max (100)
    reads_per_partition = 100



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import random
import logging
import mock
import sys
import time
import traceback
import threading
import Queue

from itertools import cycle, islice
from kazoo.testing import KazooTestCase
from threading import Event

from samsa.exceptions import NoAvailablePartitionsError
from samsa.test.integration import KafkaIntegrationTestCase, FasterKafkaIntegrationTestCase, polling_timeout
from samsa.test.case import TestCase
from samsa.cluster import Cluster
from samsa.config import ConsumerConfig
from samsa.consumer import Consumer
from samsa.topics import Topic
from samsa.partitions import Partition
from samsa.consumer.partitions import PartitionOwnerRegistry, OwnedPartition


logger = logging.getLogger(__name__)

class TestConsumerIntegration(FasterKafkaIntegrationTestCase):

    def setUp(self):
        super(TestConsumerIntegration, self).setUp()
        self.kafka = self.kafka_broker.client
        self.client.ensure_path('/brokers/topics') # can be slow to create

    def _test(self):
        if hasattr(self, 'get_topic'):
            topic = self.get_topic().name
        else:
            topic = 'testtopic'
        print 'producing for %s' % topic
        self.kafka.produce(topic, 0, ['hello', 'world'])

    def test_01(self):
        self._test()

    def test_02(self):
        self._test()

    def test_03(self):
        self._test()

    def test_04(self):
        self._test()

    def test_05(self):
        self._test()

    def test_06(self):
        self._test()

    def test_07(self):
        self._test()

    def test_08(self):
        self._test()

    def test_09(self):
        self._test()

    def test_10(self):
        self._test()

if __name__ == '__main__':
    import unittest
    unittest.main()



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import errno
import itertools
import kazoo.handlers.threading
import logging
import os
import socket
import subprocess
import tempfile
import threading
import time
import uuid

from kazoo.testing import KazooTestHarness
from nose.plugins.attrib import attr

from samsa import handlers
from samsa.client import Client
from samsa.cluster import Cluster
from samsa.test.case import TestCase
from samsa.utils.functional import methodmap


logger = logging.getLogger(__name__)


def is_port_available(port):
    """
    Checks to see if the local port is available for use.
    """
    try:
        s = socket.create_connection(('localhost', port))
        s.close()
        return False
    except IOError, err:
        return err.errno == errno.ECONNREFUSED


def merge(*dicts):
    """
    Merges a sequence of dictionaries, with values in later dictionaries
    overwriting values in the earlier dictionaries.
    """
    return reduce(lambda x, y: dict(x, **y), dicts, {})


class TimeoutError(Exception):
    pass


def polling_timeout(predicate, duration, interval=0.1,
        error='timeout exceeded'):
    """
    Provides a blocking timeout until either the predicate function returns a
    non-False value, or the timeout is exceeded in which case a `TimeoutError`
    is raised.
    """
    for _ in xrange(0, int(duration / interval)):
        if predicate():
            return
        time.sleep(interval)
    else:
        raise TimeoutError(error)


def write_property_file(properties):
    """
    Writes a dictionary if configuration properties to a file readable
    by Kafka.
    """
    file = tempfile.NamedTemporaryFile(delete=False)
    for item in properties.iteritems():
        file.write('%s=%s\n' % item)
    file.close()
    return file


class ExternalClassRunner(object):
    args = []
    kwargs = {}
    script = os.path.join(os.path.dirname(__file__), 'kafka-run-class.sh')

    stop_timeout = 3

    KAFKA_OPTIONS = {
        'heap_max': '512M',
    }

    LOGGING_CONFIGURATION = {
        'log4j.rootLogger': 'INFO, stderr',
        'log4j.appender.stderr': 'org.apache.log4j.ConsoleAppender',
        'log4j.appender.stderr.target': 'System.err',
        'log4j.appender.stderr.layout': 'org.apache.log4j.PatternLayout',
        'log4j.appender.stderr.layout.ConversionPattern': '%c: %p: %m%n',
        'log4j.logger.kafka': 'INFO',
    }

    def __init__(self):
        self.logging_configuration_file = \
            write_property_file(self.LOGGING_CONFIGURATION)

    def start(self):
        args = ['/bin/sh', self.script, self.cls] + self.args

        kwargs = self.kwargs.copy()
        kwargs['env'] = kwargs.get('env', os.environ).copy()
        kwargs['env'].update({
            'KAFKA_OPTS': '-Xmx%(heap_max)s -server '
                '-Dlog4j.configuration=file:%(logging_config)s' %
                merge(self.KAFKA_OPTIONS, {
                    'logging_config': self.logging_configuration_file.name,
                })
        })

        self.process = subprocess.Popen(args, stderr=subprocess.PIPE, **kwargs)

        def convert_log_output(namespace):
            """
            Scrapes log4j output, forwarding the log output to the
            corresponding Python logging endpoint.
            """
            while True:
                line = self.process.stderr.readline().strip()
                if not line:
                    continue

                try:
                    name, level, message = \
                        methodmap('strip', line.split(':', 2))
                    logger = logging.getLogger('%s.%s' % (namespace, name))
                    logger.log(getattr(logging, level.upper()), message)
                except Exception:
                    logger = logging.getLogger('%s.raw' % namespace)
                    logger.warning(line)

        self.log_thread = threading.Thread(target=convert_log_output,
            args=('java.%s' % self.cls,))
        self.log_thread.daemon = True  # shouldn't be necessary, just in case
        self.log_thread.start()

    def is_running(self):
        if not hasattr(self, 'process'):
            return False

        self.process.poll()
        return self.process.returncode is None

    def stop(self, timeout=None):
        if not self.is_running():
            return

        if timeout is None:
            timeout = self.stop_timeout

        logger.debug('Sending SIGTERM to %s...', self.process)

        self.process.terminate()
        # Can block, but we're reading from the pipe so it should be fine.
        self.process.wait()


class ManagedBroker(ExternalClassRunner):
    cls = 'kafka.Kafka'

    start_timeout = int(os.environ.get('KAFKA_START_TIMEOUT', 3))

    CONFIGURATION = {
        'enable.zookeeper': 'true',
        'log.flush.interval': 1,  # commit messages as soon as possible
    }

    def __init__(self, zookeeper, hosts, brokerid=0, partitions=1, port=9092):
        super(ManagedBroker, self).__init__()
        self.directory = tempfile.mkdtemp()

        self.zookeeper = zookeeper
        self.brokerid = brokerid
        self.partitions = partitions
        self.port = port

        self.configuration = merge(self.CONFIGURATION, {
            'log.dir': self.directory,
            'zk.connect': hosts,
            'brokerid': self.brokerid,
            'num.partitions': self.partitions,
            'hostname': '127.0.0.1', # travis-ci will try to make this ipv6
            'port': self.port,
        })

        self.configuration_file = write_property_file(self.configuration)
        self.args = [self.configuration_file.name]
        self._client = None

    def start(self, timeout=None):
        if timeout is None:
            timeout = self.start_timeout

        ready = threading.Event()
        path = '/brokers/ids/%s' % self.brokerid
        callback = lambda *a, **k: ready.set()
        if self.zookeeper.exists(path, watch=callback) is not None:
            raise AssertionError('Kafka broker with broker ID %s is already '
                'running!' % self.brokerid)

        super(ManagedBroker, self).start()

        for _ in xrange(0, timeout):
            if ready.is_set():
                break
            ready.wait(1)
        else:
            raise TimeoutError('Kafka broker did not start within %s seconds'
                % timeout)

    def stop(self, *args, **kwargs):
        logger.debug('Shutting down Kafka broker...')
        super(ManagedBroker, self).stop(*args, **kwargs)
        # TODO: Remove configuration, log dir

    @property
    def client(self):
        assert self.is_running()
        if not self._client:
            self._client = Client(
                'localhost', handlers.ThreadingHandler(),
                port=self.port
            )
        return self._client


class ManagedProducer(ExternalClassRunner):
    cls = 'kafka.tools.ProducerShell'
    kwargs = {
        'stdout': open('/dev/null'),
        'stdin': subprocess.PIPE,
    }

    CONFIGURATION = {
        'producer.type': 'sync',
        'compression.codec': 0,
        'serializer.class': 'kafka.serializer.StringEncoder',
    }

    def __init__(self, hosts, topic):
        super(ManagedProducer, self).__init__()
        self.hosts = hosts
        self.topic = topic

        self.configuration_file = write_property_file(merge(
            self.CONFIGURATION, {
                'zk.connect': self.hosts,
            }))

        self.args = ['--topic', self.topic,
            '--props', self.configuration_file.name]

    def publish(self, messages):
        stream = self.process.stdin
        for message in messages:
            stream.write('%s\n' % message)
        stream.flush()


class ManagedConsumer(ExternalClassRunner):
    cls = 'kafka.tools.ConsumerShell'
    kwargs = {
        'stdout': subprocess.PIPE,
    }

    def __init__(self, hosts, topic, group='test-consumer-group'):
        super(ManagedConsumer, self).__init__()
        self.hosts = hosts
        self.topic = topic
        self.group = group

        self.configuration_file = write_property_file({
            'zk.connect': self.hosts,
            'groupid': self.group,
        })

        self.args = ['--topic', self.topic,
            '--props', self.configuration_file.name]


@attr('integration')
class KafkaIntegrationTestCase(TestCase, KazooTestHarness):
    """
    A test case that allows the bootstrapping of a number of Kafka brokers.

    Brokers are restarted with new log directories for each test method
    invocation, ensuring that individual tests do not affect the results of
    others.

    All managed subprocesses (brokers, consumers, producers, etc.) are
    automatically stopped when the test case is torn down.
    """

    def setUp(self):
        """Set up kafka and zookeeper."""
        try:
            self.setup_zookeeper()
        except kazoo.handlers.threading.TimeoutError:
            logging.warning('Zookeeper failed to start. Trying again.')
            if self.cluster[0].running:
                self.cluster.stop()
            self.setup_zookeeper() # try again if travis-ci is being slow

        # Keep track of processes started
        self._consumers = []
        self._producers = []

        self.kafka_broker = self.start_broker(self.client, self.hosts)
        self.kafka_cluster = Cluster(self.client)

    def tearDown(self):
        """Reset zookeeper and Kafka if needed"""
        self.client.stop()
        for process in itertools.chain(self._consumers, self._producers):
            process.stop()
        if self.kafka_broker and self.kafka_broker.is_running():
            self.kafka_broker.stop()
        self.teardown_zookeeper()

    @classmethod
    def start_broker(cls, zk_client, zk_hosts, brokerid=0):
        ports = itertools.ifilter(
            is_port_available, itertools.count(9092)
        )
        broker = ManagedBroker(
            zk_client,
            zk_hosts,
            brokerid=brokerid,
            port=next(ports),
        )
        broker.start()
        return broker

    def consumer(self, *args, **kwargs):
        consumer = ManagedConsumer(self.hosts, *args, **kwargs)
        consumer.start()
        self._consumers.append(consumer)
        return consumer

    def producer(self, *args, **kwargs):
        producer = ManagedProducer(self.hosts, *args, **kwargs)
        producer.start()
        self._producers.append(producer)
        return producer

    def get_topic(self):
        """Get a random topic to use for testing."""
        topic = str(uuid.uuid4())
        return self.kafka_cluster.topics[topic]


class FasterKafkaIntegrationTestCase(KafkaIntegrationTestCase):
    """A faster test case that doesn't restart Kafka between tests.

    To use this class, use `self.get_topic()` instead of hard-coding
    the topic name. This lets the same broker be used for all tests
    without having to worry about collisions between tests.

    This exists because kafka can't delete topics, requiring a full
    restart between test cases. This is very slow and not strictly
    necessary for most tests.

    If your test *needs* the broker to restart between cases, or you're
    testing the broker going down, etc, then use `KafkaIntegrationTestCase`.
    However, for most tests, this is the class you want.
    """

    @classmethod
    def setUpClass(cls):
        cls.zk_harness = KazooTestHarness()
        cls.zk_harness.setup_zookeeper()
        cls.client = cls.zk_harness.client
        cls.hosts = cls.zk_harness.hosts
        cls.kafka_broker = cls.start_broker(cls.client, cls.hosts)
        cls.kafka_cluster = Cluster(cls.client)

    @classmethod
    def tearDownClass(cls):
        cls.client.stop()
        if cls.kafka_broker and cls.kafka_broker.is_running():
            cls.kafka_broker.stop()
        cls.zk_harness.teardown_zookeeper()

    def setUp(self):
        # Keep track of processes started
        self._consumers = []
        self._producers = []

    def tearDown(self):
        """Only need to stop consumer/producer processes"""
        for process in itertools.chain(self._consumers, self._producers):
            process.stop()



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import mock
import sys
import time
import traceback
import threading
import Queue

from itertools import cycle, islice
from kazoo.testing import KazooTestCase
from threading import Event

from samsa.exceptions import NoAvailablePartitionsError
from samsa.test.integration import KafkaIntegrationTestCase, polling_timeout
from samsa.test.integration import FasterKafkaIntegrationTestCase, polling_timeout
from samsa.test.case import TestCase
from samsa.cluster import Cluster
from samsa.config import ConsumerConfig
from samsa.consumer import Consumer
from samsa.topics import Topic
from samsa.partitions import Partition
from samsa.consumer.partitions import PartitionOwnerRegistry, OwnedPartition


logger = logging.getLogger(__name__)

class TestPartitionOwnerRegistry(KazooTestCase):
    """Test the methods of :class:`samsa.consumer.PartitionOwnerRegistry`.
    """

    @mock.patch('samsa.cluster.BrokerMap')
    def setUp(self, bm, *args):
        super(TestPartitionOwnerRegistry, self).setUp()
        self.c = Cluster(self.client)
        broker = mock.Mock()
        broker.id = 1
        self.c.brokers.__getitem__.return_value = broker

        self.consumer = mock.Mock()
        self.consumer.id  = '1234'
        self.topic = mock.Mock()
        self.topic.name = 'topic'

        self.por = PartitionOwnerRegistry(
            self.consumer,
            self.c,
            self.topic,
            'group'
        )

        # Create 5 partitions with on the same topic and broker
        self.partitions = []
        for i in xrange(5):
            self.partitions.append(
                OwnedPartition(Partition(self.c, self.topic, broker, i), 'group')
            )

    @mock.patch.object(OwnedPartition, 'start')
    def test_crd(self, *args):
        """Test partition *c*reate, *r*ead, and *d*elete.
        """

        # Add the first 3 partitions to the registry and see that they're set
        self.por.add(self.partitions[:3])
        self.assertEquals(
            self.por.get(),
            set(self.partitions[:3])
        )

        # Remove the first partition and see that only [1, 2] exist
        self.por.remove([self.partitions[0]])
        self.assertEquals(
            self.por.get(),
            set(self.partitions[1:3])
        )

    @mock.patch.object(OwnedPartition, 'start')
    def test_grows(self, *args):
        """Test that the reference returned by
        :func:`samsa.consumer.partitions.PartitionOwnerRegistry.get` reflects
        the latest state.

        """
        partitions = self.por.get()
        self.assertEquals(len(partitions), 0)

        self.por.add(self.partitions)
        self.assertEquals(len(partitions), len(self.partitions))


class TestConsumer(KazooTestCase, TestCase):

    def setUp(self):
        super(TestConsumer, self).setUp()
        self.client.ensure_path("/brokers/ids")
        self.client.ensure_path('/brokers/topics')
        self.c = Cluster(self.client)

    def tearDown(self):
        self.client.stop() # stops watches that call _rebalance
        super(TestConsumer, self).tearDown()

    def _register_fake_brokers(self, n=1, client=None):
        if client is None:
            client = self.client
        client.ensure_path("/brokers/ids")
        for i in xrange(n):
            data = "creator:127.0.0.1:%s" % (9092 + i)
            self._register_fake_broker(i, data, client=client)

    def _register_fake_broker(self, name, data, client=None):
        if client is None:
            client = self.client
        path = "/brokers/ids/%s" % name
        client.create(path, data)

    def _register_fake_partitions(self, topic, n_partitions=1, brokers=None, client=None):
        """Register fake partitions. Specify broker list or use None to create for all"""
        if client is None:
            client = self.client
        brokers = brokers or client.get_children("/brokers/ids")
        for broker in brokers:
            client.ensure_path("/brokers/topics/%s" % topic)
            part_path = "/brokers/topics/%s/%s" % (topic, broker)
            client.create(part_path, str(n_partitions))


    @mock.patch.object(OwnedPartition, 'start')
    def test_assigns_partitions(self, *args):
        """
        Test rebalance

        Adjust n_* to see how rebalancing performs.

        """
        n_partitions = 10
        n_consumers = 3
        self._register_fake_brokers(n_partitions)
        t = Topic(self.c, 'testtopic')

        consumers = [t.subscribe('group1') for i in xrange(n_consumers)]

        partitions = []
        for c in consumers:
            partitions.extend(c.partitions)

        # test that there are no duplicates.
        self.assertEquals(len(partitions), n_partitions)
        # test that every partitions is represented.
        self.assertEquals(len(set(partitions)), n_partitions)


    @mock.patch.object(OwnedPartition, 'start')
    def test_broker_addition(self, rebalance):
        """Test adding a broker, and ensure all partitions are discovered

        Testing this makes sure all the relevant zookeeper watches are
        functioning and rebalancing is happening when partitions or brokers
        are added or removed
        """
        t = Topic(self.c, 'testtopic')

        self._register_fake_broker(0, "creator:127.0.0.1:9092")
        self._register_fake_partitions('testtopic', n_partitions=2)

        consumer = t.subscribe('group1')

        self._register_fake_broker(1, "creator:127.0.0.1:9093")
        self._register_fake_partitions('testtopic', n_partitions=2, brokers=['1'])

        time.sleep(1) # let watches resolve
        self.assertEqual(len(t.partitions), 4)
        self.assertEqual(len(consumer.partitions), 4)


    @mock.patch.object(Partition, 'fetch')
    def test_commits_offsets(self, fetch):
        """Test that message offsets are persisted to ZK.

        """
        self._register_fake_brokers(1)
        t = Topic(self.c, 'testtopic')

        c = t.subscribe('group')
        msgs = []
        for i in xrange(1, 10):
            msg = mock.Mock()
            msg.next_offset = 3 * i
            msg.payload = str(i) * 3
            msgs.append(msg)
        fetch.return_value = msgs

        self.assertPassesWithMultipleAttempts(
            lambda: self.assertTrue(c.next_message(10) is not None),
            5
        )
        self.assertEquals(len(c.partitions), 1)
        p = list(c.partitions)[0]

        self.assertEquals(p.offset, 3)
        c.commit_offsets()

        d, stat = self.client.get(p.path)
        self.assertEquals(d, '3')
        c.stop_partitions()

    @mock.patch.object(Partition, 'fetch')
    def test_consumer_remembers_offset(self, fetch):
        """Test that offsets are successfully retrieved from zk.

        """
        topic = 'testtopic'
        group = 'testgroup'
        offset = 10
        ev = Event()

        fake_partition = mock.Mock()
        fake_partition.cluster = self.c
        fake_partition.topic.name = topic
        fake_partition.broker.id = 0
        fake_partition.number = 0
        def fake_fetch(*args, **kwargs):
            ev.set()
            return ()
        fetch.side_effect = fake_fetch

        op = OwnedPartition(fake_partition, group)
        op._current_offset = offset
        op.commit_offset()

        self._register_fake_brokers(1)
        t = Topic(self.c, topic)
        c = t.subscribe(group)

        self.assertEquals(len(c.partitions), 1)
        p = list(c.partitions)[0]
        self.assertEquals(p.offset, offset)

        self.assertEquals(None, p.next_message(0))
        ev.wait(1)
        fetch.assert_called_with(offset, ConsumerConfig.fetch_size)
        c.stop_partitions()


    @mock.patch.object(OwnedPartition, 'start')
    def test_multiclient_rebalance(self, *args):
        """Test rebalancing with many connected clients

        This test is primarily good at ferreting out concurrency bugs
        and therefore doesn't test one specific thing, since such
        bugs are fundamentally hard to trap.

        To test, it simulates 10 consumer connecting over time, rebalancing,
        and eventually disconnecting one by one. In order to accomplish
        this, it creates a separate kazoo client for each consumer. This
        is required because otherwise there is too much thread contention
        over the single kazoo client. Some callbacks won't happen until
        much too late because there aren't enough threads to go around.

        """
        n_partitions = 10
        n_consumers = 10
        t = Topic(self.c, 'testtopic')

        # with self.client, new clients don't see the brokers -- unsure why
        from kazoo.client import KazooClient
        zk_hosts = ','.join(
                ['%s:%s' % (h,p) for h,p in self.client.hosts.hosts])
        zkclient = KazooClient(hosts=zk_hosts)
        zkclient.start()
        self._register_fake_brokers(n_partitions, client=zkclient)
        zkclient.ensure_path('/brokers/topics')

        # bring up consumers
        consumers = []
        for i in xrange(n_consumers):
            newclient = KazooClient(hosts=zk_hosts)
            newclient.start()
            cluster = Cluster(newclient)
            topic = Topic(cluster, 'testtopic')
            consumers.append((newclient, topic.subscribe('group1')))
            time.sleep(1)

        time.sleep(5) # let things settle

        # bring down consumers
        for client,consumer in consumers:
            consumer.stop_partitions()
            client.stop()
            time.sleep(2) # a little more time so we don't kill during rebalance

        newclient.stop()

    @mock.patch.object(OwnedPartition, 'start')
    def test_too_many_consumers(self, *args):
        """Test graceful failure when # of consumers exceeds partitions

        """
        n_partitions = 1
        n_consumers = 2
        self._register_fake_brokers(n_partitions)
        t = Topic(self.c, 'testtopic')

        with self.assertRaises(NoAvailablePartitionsError):
            consumers = [t.subscribe('group1') for i in xrange(n_consumers)]


class TestConsumerIntegration(FasterKafkaIntegrationTestCase):

    def setUp(self):
        super(TestConsumerIntegration, self).setUp()
        self.kafka = self.kafka_broker.client
        self.client.ensure_path('/brokers/topics') # can be slow to create


    def test_consumes(self):
        """Test that we can consume messages from kafka.

        """
        topic = self.get_topic().name
        messages = ['hello world', 'foobar']

        # publish `messages` to topic
        self.kafka.produce(topic, 0, messages)

        t = Topic(self.kafka_cluster, topic)

        # subscribe to topic
        consumer = t.subscribe('group2')

        def test():
            """Test that `consumer` can see `messages`.

            catches exceptions so we can retry while we wait for kafka to
            coallesce.

            """
            logger.debug('Running `test`...')
            try:
                self.assertEquals(
                    list(islice(consumer, 0, len(messages))),
                    messages
                )
                return True
            except AssertionError as e:
                logger.exception('Caught exception: %s', e)
                return False

        # wait for one second for :func:`test` to return true or raise an error
        polling_timeout(test, 1)

        old_offset = [p.offset for p in consumer.partitions][0]
        # test that the offset of our 1 partition is not 0
        self.assertTrue(old_offset > 0)
        # and that consumer contains no more messages.
        self.assertTrue(consumer.empty())

        # repeat and see if offset grows.
        self.kafka.produce(topic, 0, messages)
        polling_timeout(test, 1)
        self.assertTrue([p.offset for p in consumer.partitions][0] > old_offset)
        consumer.stop_partitions()


    def test_empty_topic(self):
        """Test that consuming an empty topic returns an empty list.

        """
        topic = self.get_topic().name
        t = Topic(self.kafka_cluster, topic)

        consumer = t.subscribe('group2')
        self.assertTrue(consumer.empty())
        consumer.stop_partitions()


    def test_fetch_invalid_offset(self):
        """Test that fetching rolled-over offset skips to lowest valid offset.

        Bad offsets happen when kafka logs are rolled over automatically.
        This could happen when a consumer is offline for a long time and
        zookeeper has an old offset stored. It could also happen with a
        consumer near the end of a log that's being rolled over and its
        previous spot no longer exists.
        """
        topic = self.get_topic().name
        messages = ['hello world', 'foobar']

        # publish `messages` to topic
        self.kafka.produce(topic, 0, messages)

        t = Topic(self.kafka_cluster, topic)

        # get the consumer and set the offset to -1
        consumer = t.subscribe('group2')
        list(consumer.partitions)[0]._next_offset = -1

        def test():
            """Test that `consumer` can see `messages`.

            catches exceptions so we can retry while we wait for kafka to
            coallesce.

            """
            logger.debug('Running `test`...')
            try:
                self.assertEquals(
                    list(islice(consumer, 0, len(messages))),
                    messages
                )
                return True
            except AssertionError as e:
                logger.exception('Caught exception: %s', e)
                return False

        # wait for one second for :func:`test` to return true or raise an error
        polling_timeout(test, 1)
        consumer.stop_partitions()



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from samsa.handlers import ThreadingHandler
from samsa.brokers import BrokerMap
from samsa.topics import TopicMap


class Cluster(object):
    """
    A Kafka cluster.

    :ivar brokers: The :class:`samsa.brokers.BrokerMap` for this cluster.
    :ivar topics: The :class:`samsa.topics.TopicMap` for this cluster.

    :param zookeeper: A ZooKeeper client.
    :type zookeeper: :class:`kazoo.client.Client`
    :param handler: Async handler.
    :type handler: :class:`kazoo.handlers.Handler`
    """
    def __init__(self, zookeeper, handler=None):
        if not zookeeper.connected:
            raise Exception("Zookeeper must be connected before use.")

        self.zookeeper = zookeeper

        if not handler:
            handler = ThreadingHandler()
        self.handler = handler

        self.brokers = BrokerMap(self)
        self.topics = TopicMap(self)



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import itertools
import logging
import random
import string
import struct
import sys
from zlib import crc32

import unittest2

from samsa.client import (Client, Message, OFFSET_EARLIEST, OFFSET_LATEST,
    COMPRESSION_TYPE_GZIP)
from samsa.exceptions import InvalidVersionError, WrongPartitionError
from samsa.test.integration import FasterKafkaIntegrationTestCase
from samsa.test.integration import KafkaIntegrationTestCase


logger = logging.getLogger(__name__)

# class ClientTestCase(unittest2.TestCase):
#     def test_produce(self):
#         raise NotImplementedError
#
#     def test_multiproduce(self):
#         raise NotImplementedError
#
#     def test_fetch(self):
#         raise NotImplementedError
#
#     def test_multifetch(self):
#         raise NotImplementedError
#
#     def test_offsets(self):
#         raise NotImplementedError


class MessageTestCase(unittest2.TestCase):
    def setUp(self):
        self.payload = 'hello world'
        self.valid_checksum = crc32(self.payload)

    def test_06_format(self):
        magic = 0

        def make_message(payload, checksum=None):
            if checksum is None:
                checksum = self.valid_checksum
            encoded = ''.join([struct.pack('!bi', magic, checksum), payload])
            framed = buffer(''.join([struct.pack('!i', len(encoded)),
                encoded]))
            return Message(framed)

        message = make_message(self.payload)
        self.assertTrue(message.valid)
        self.assertEqual(message.payload, self.payload)
        self.assertDictContainsSubset({
            'checksum': self.valid_checksum,
            'magic': magic,
        }, message.headers)

        self.assertFalse(make_message(self.payload,
            self.valid_checksum - 1).valid)

    def test_07_format(self):
        magic = 1

        def make_message(payload, checksum=None, compression=0):
            if checksum is None:
                checksum = self.valid_checksum
            encoded = ''.join([
                struct.pack('!bbi', magic, compression, checksum), payload])
            framed = buffer(''.join([struct.pack('!i', len(encoded)),
                                     encoded]))
            return Message(framed)

        message = make_message(self.payload)
        self.assertTrue(message.valid)
        self.assertEqual(message.payload, self.payload)
        self.assertDictContainsSubset({
            'checksum': self.valid_checksum,
            'magic': magic,
        }, message.headers)

        # TODO: Test with compression

        self.assertFalse(make_message(self.payload,
            self.valid_checksum - 1).valid)


def filter_messages(stream):
    # TODO: this can deadlock, needs a timeout or something
    prefix = 'consumed: '
    while True:
        line = stream.readline().strip()
        if line.startswith(prefix):
            message = line[len(prefix):]
            yield message


class ClientIntegrationTestCase(KafkaIntegrationTestCase):
    def setUp(self):
        super(ClientIntegrationTestCase, self).setUp()
        self.kafka = self.kafka_broker.client

    def test_produce(self, count=1, **kwargs):
        topic = self.get_topic().name
        message = 'hello world'
        consumer = self.consumer(topic)
        messages = (message,) * count
        self.kafka.produce(topic, 0, messages, **kwargs)

        consumed = list(itertools.islice(filter_messages(consumer.process.stdout), count))
        self.assertEqual(consumed, list(messages))

    def test_produce_with_multiple_messages(self):
        self.test_produce(count=3)

    def test_produce_version_1_no_compression(self):
        self.test_produce(version=1)

    def test_produce_invalid_compression(self):
        with self.assertRaises(ValueError):
            self.test_produce(version=1, compression=sys.maxint)

    def test_produce_invalid_version_for_compression(self):
        with self.assertRaises(ValueError):
            self.test_produce(version=0, compression=COMPRESSION_TYPE_GZIP)

    def test_produce_invalid_version(self):
        with self.assertRaises(InvalidVersionError):
            self.kafka.produce('topic', 0, ('hello world',), version=sys.maxint)

    def test_multiproduce(self, count=1, **kwargs):
        topics = (self.get_topic().name, self.get_topic().name)

        def message_for_topic(topic):
            return 'hello to topic %s' % topic

        consumers = {}
        for topic in topics:
            consumer = self.consumer(topic)
            consumers[topic] = consumer

        batch = []
        for topic in topics:
            batch.append((topic, 0, (message_for_topic(topic),) * count))

        self.kafka.multiproduce(batch, **kwargs)

        for topic, consumer in consumers.items():
            consumed = list(itertools.islice(filter_messages(consumer.process.stdout), count))
            messages = list((message_for_topic(topic),) * count)
            self.assertEqual(consumed, messages)

    def test_multiproduce_with_multiple_messages(self):
        self.test_multiproduce(count=3)

    def test_multiproduce_version_1_no_compression(self):
        self.test_multiproduce(version=1)

    def test_multiproduce_invalid_version(self):
        topic = self.get_topic().name
        with self.assertRaises(InvalidVersionError):
            self.kafka.multiproduce((
                (topic, 0, ('hello world',)),
            ), version=sys.maxint)

    def test_fetch(self):
        # TODO: test error conditions
        topic = self.get_topic().name
        payload = 'hello world'
        size = 1024 * 300

        producer = self.producer(topic)
        producer.publish([payload])

        def ensure_valid_response():
            messages = list(self.kafka.fetch(topic, 0, 0, size))
            self.assertEqual(len(messages), 1)

            message = messages[0]
            self.assertIsInstance(message, Message)
            self.assertEqual(message.offset, 0)
            self.assertEqual(message.next_offset, len(message))
            self.assertEqual(message.payload, payload)
            self.assertEqual(message['compression'], 0)
            self.assertTrue(message.valid)

            self.offset = message.next_offset

        self.assertPassesWithMultipleAttempts(
            ensure_valid_response, 5,
            backoff=lambda attempt, timeout: (
                timeout * sum(xrange(1, attempt + 1))
            )
        )

        payloads = ['hello', 'world']
        producer.publish(payloads)

        def ensure_valid_response_again():
            messages = list(self.kafka.fetch(topic, 0, self.offset, size))
            self.assertEqual(len(messages), 2)
            self.assertTrue(all(isinstance(m, Message) for m in messages))
            self.assertEqual([m.payload for m in messages], payloads)
            self.assertEqual(messages[0].offset, self.offset)

        self.assertPassesWithMultipleAttempts(ensure_valid_response_again, 5)

    def test_fetch_sizing(self):
        topic = self.get_topic().name
        partition = 0
        payload = ''.join(random.choice(string.ascii_letters) for _
            in xrange(0, 300))

        producer = self.producer(topic)
        producer.publish([payload])

        def ensure_no_partial_messages():
            size = len(payload) // 2
            messages = list(self.kafka.fetch(topic, partition, 0, size))
            self.assertEqual(len(messages), 0)

            messages = list(self.kafka.fetch(topic, partition, 0, 1024 * 300))
            self.assertEqual(len(messages), 1)
            message = messages[0]
            self.assertTrue(message.valid)
            self.assertEqual(message.payload, payload)

        self.assertPassesWithMultipleAttempts(ensure_no_partial_messages, 5)

    def test_fetch_wrong_partition(self):
        topic = self.get_topic().name
        with self.assertRaises(WrongPartitionError):
            self.kafka.fetch(topic, 10, 0, 1024 * 300)

    def test_multifetch(self):
        # TODO: test error conditions
        # In this case, sorting matters because of how data comes back
        topics = (self.get_topic().name, self.get_topic().name)
        size = 1024 * 300

        def payload_for_topic(topic):
            return 'hello from topic %s' % topic

        producers = {}
        for topic in topics:
            producer = self.producer(topic)
            producer.publish([payload_for_topic(topic)])
            producers[topic] = producer

        def ensure_valid_response():
            batches = [(topic, 0, 0, size) for topic in topics]
            responses = self.kafka.multifetch(batches)

            self.next_offsets = {}
            num_responses = 0
            for topic, response in zip(topics, responses):
                messages = list(response)
                self.assertEqual(len(messages), 1)

                message = messages[0]
                self.assertIsInstance(message, Message)
                self.assertEqual(message.offset, 0)
                self.assertEqual(message.next_offset, len(message))
                self.next_offsets[topic] = message.next_offset
                self.assertEqual(message.payload, payload_for_topic(topic))
                self.assertEqual(message['compression'], 0)
                self.assertTrue(message.valid)

                num_responses += 1

            self.assertEqual(len(batches), num_responses)

        self.assertPassesWithMultipleAttempts(ensure_valid_response, 5)

        batches = []
        num_messages = 2
        for topic in topics:
            payloads = [payload_for_topic(topic)] * num_messages
            producers[topic].publish(payloads)
            batches.append((topic, 0, self.next_offsets[topic], size))

        def ensure_valid_response_again():
            responses = self.kafka.multifetch(batches)
            for topic, response in zip(topics, responses):
                messages = list(response)
                self.assertEqual(len(messages), num_messages)

                message = messages[0]
                self.assertEqual(message.offset, self.next_offsets[topic])
                self.assertEqual([m.payload for m in messages],
                    [payload_for_topic(topic)] * num_messages)

        self.assertPassesWithMultipleAttempts(ensure_valid_response_again, 5)

    def test_offsets(self):
        topic = self.get_topic().name

        offsets = self.kafka.offsets(topic, 0, OFFSET_EARLIEST, 1)
        self.assertEqual(len(offsets), 1)
        self.assertEqual(offsets[0], 0)

        offsets = self.kafka.offsets(topic, 0, OFFSET_LATEST, 1)
        self.assertEqual(len(offsets), 1)
        self.assertEqual(offsets[0], 0)

    def test_bad_offset(self):
        """Test recovering from a bad offset stored in Zookeeper"""
        topic = 'topic'
        payload = 'hello world'
        size = 1024 * 300

        producer = self.producer(topic)
        producer.publish([payload]*5)

        # We don't have 100% recovery, so test at offsets that work
        offsets = ((0,5), (2,4), (8,4), (25,3), (46,2), (67,1))
        def ensure_recovery():
            for offset,len_ in offsets:
                messages = list(self.kafka.fetch(topic, 0, offset, size))
                self.assertEqual(len(messages), len_)
                continue

        self.assertPassesWithMultipleAttempts(
            ensure_recovery, 5,
            backoff=lambda attempt, timeout: (
                timeout * sum(xrange(1, attempt + 1))
            )
        )



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import socket
import struct
from zlib import crc32

from samsa import handlers
from samsa.exceptions import (ERROR_CODES, InvalidVersionError,
    SocketDisconnectedError)
from samsa.utils import attribute_repr
from samsa.utils.functional import methodimap
from samsa.utils.namedstruct import NamedStruct
from samsa.utils.socket import recv_framed
from samsa.utils.structuredio import StructuredBytesIO


logger = logging.getLogger(__name__)


# Requests

(REQUEST_TYPE_PRODUCE, REQUEST_TYPE_FETCH, REQUEST_TYPE_MULTIFETCH,
    REQUEST_TYPE_MULTIPRODUCE, REQUEST_TYPE_OFFSETS) = range(0, 5)

OFFSET_LATEST = -1
OFFSET_EARLIEST = -2

DEFAULT_VERSION = 0

COMPRESSION_TYPE_NONE = 0
COMPRESSION_TYPE_GZIP = 1
COMPRESSION_TYPE_SNAPPY = 2
COMPRESSION_TYPES = (
    COMPRESSION_TYPE_NONE, COMPRESSION_TYPE_GZIP,
    COMPRESSION_TYPE_SNAPPY
)

MessageSetFrameHeader = NamedStruct('MessageSetFrameHeader', (
    ('i', 'length'),
))

# Responses

MessageSetHeader = NamedStruct('MessageSetHeader', (
    ('i', 'length'),
    ('h', 'error'),
))

ResponseFrameHeader = struct.Struct('!i')

ResponseErrorHeader = NamedStruct('ResponseErrorHeader', (
    ('h', 'error'),
))

OffsetsResponseHeader = NamedStruct('OffsetsResponseHeader', (
    ('i', 'count'),
))

Offset = NamedStruct('Offset', (
    ('q', 'value'),
))


# Messages

class VersionHeaderMap(dict):
    def __missing__(self, key):
        raise InvalidVersionError('%s is not a valid version identifier' % key)


class Message(object):
    __slots__ = ('_headers', 'raw', 'offset')

    Header = NamedStruct('Header', (
        ('i', 'length'),
        ('b', 'magic'),
    ))

    VersionHeaders = VersionHeaderMap({
        0: NamedStruct('Header', (
            ('i', 'checksum'),
        )),
        1: NamedStruct('HeaderWithCompression', (
            ('b', 'compression'),
            ('i', 'checksum'),
        )),
    })

    def __init__(self, raw, offset=0):
        self.raw = raw
        self.offset = offset

        self._headers = []
        header = self.Header.unpack_from(self.raw)
        self._headers.append(header)

        versioned_header = self.VersionHeaders[header.magic].unpack_from(
            self.raw, offset=self.Header.size
        )
        self._headers.append(versioned_header)

    __repr__ = attribute_repr('raw', 'offset')

    def __len__(self):
        return len(self.raw)

    def __str__(self):
        return str(self.payload)

    def __getitem__(self, name):
        for header in self._headers:
            try:
                return getattr(header, name)
            except AttributeError:
                pass
        else:
            raise AttributeError('%s does not have a field named "%s".' % (
                repr(self), name)
            )

    @property
    def headers(self):
        return reduce(
            lambda x, y: dict(x, **y),
            methodimap('_asdict', self._headers), {}
        )

    def get(self, name, default=None):
        try:
            return self[name]
        except AttributeError:
            return default

    @property
    def next_offset(self):
        return self.offset + len(self)

    @property
    def payload(self):
        start = self.Header.size + self.VersionHeaders[self['magic']].size
        return self.raw[start:]

    @property
    def valid(self):
        return self['checksum'] == crc32(self.payload)

    @classmethod
    def pack_into(cls, bytea, offset, payload, version, compression=None):
        """
        Packs a message payload into a buffer.

        :param bytea: buffer to pack the message into
        :type bytea: bytearray
        :param offset: offset to start writing at
        :type offset: int
        :param payload: message payload
        :type payload: str
        :param version: message version to publish ("magic number")
        :type version: int
        :param compression: which compression format to use
                            (only for version 1)
        :type compression: int
        :returns: total amount of written (in bytes)
        :rtype: int
        """
        if version < 1:
            if compression is not None:
                raise ValueError(
                    'Compression is not supported on version %s' % version
                )
        elif compression is not None:
            if compression not in COMPRESSION_TYPES:
                raise ValueError(
                    '%s is not a valid compression type' % compression
                )
            elif compression is not COMPRESSION_TYPE_NONE:
                raise NotImplementedError  # TODO
        else:
            compression = COMPRESSION_TYPE_NONE

        VersionHeader = cls.VersionHeaders[version]

        # Write generic message header.
        length = cls.Header.size + VersionHeader.size + len(payload)
        cls.Header.pack_into(
            bytea, offset=offset,
            length=length - 4, magic=version
        )
        offset += cls.Header.size

        # Write versioned message header.
        version_kwargs = {}
        if compression is not None:
            version_kwargs['compression'] = compression
        VersionHeader.pack_into(
            bytea, offset=offset,
            checksum=crc32(payload), **version_kwargs
        )
        offset += VersionHeader.size

        # Write message payload.
        bytea[offset:offset + len(payload)] = payload

        return length

    @classmethod
    def encode(cls, messages, version, **kwargs):
        """
        Encodes multiple messages.

        :param messages: messages to publish
        :type messages: sequence of strs
        :param version: message version to publish ("magic number")
        :type version: int
        :param \*\*kwargs: extra arguments to pass to :meth:`.pack_into`
        :returns: encoded messages
        :rtype: :class:`samsa.utils.structuredio.StructuredBytesIO`
        """
        message_header_length = (
            cls.Header.size +
            cls.VersionHeaders[version].size
        )
        length = (
            MessageSetFrameHeader.size +
            sum(map(len, messages)) +
            (len(messages) * message_header_length)
        )
        bytea = bytearray(length)
        MessageSetFrameHeader.pack_into(bytea, 0, length=length - 4)
        offset = MessageSetFrameHeader.size
        for message in messages:
            written = cls.pack_into(
                bytea, offset,
                payload=message, version=version, **kwargs
            )
            offset += written
        return StructuredBytesIO(bytea)


def decode_message_sets(payload, from_offsets):
    offset = 0
    for from_offset in from_offsets:
        (length,) = ResponseFrameHeader.unpack_from(payload, offset=offset)
        header = ResponseErrorHeader.unpack_from(
            payload,
            offset=offset + ResponseFrameHeader.size
        )
        if header.error:
            error_class = ERROR_CODES.get(header.error, -1)
            raise error_class(error_class.reason)
        message_set_payload = buffer(payload,
            offset + (ResponseFrameHeader.size + ResponseErrorHeader.size),
            length - ResponseErrorHeader.size)
        yield decode_messages(message_set_payload, from_offset)
        offset += length + ResponseFrameHeader.size


def decode_messages(payload, from_offset):
    """
    Decodes ``Message`` objects from a ``payload`` buffer.
    """
    offset = 0
    recovering = False # recovering from bad offset error
    while offset < len(payload):
        message = None
        try:
            header = Message.Header.unpack_from(payload, offset)
            length = 4 + header.length
            if length > 0:
                message = Message(
                    raw=buffer(payload, offset, length),
                    offset=from_offset + offset
                )
        except struct.error:
            if not recovering:
                # Thrown if payload ends in the middle of a header
                logger.debug('Unable to create message from payload remainder '
                             '%i bytes left: %s',
                             len(payload) - offset, payload[offset:])
                return
        except InvalidVersionError, ex:
            if not recovering: # enter recovery mode to find next valid message
                recovering = True
                exception = ex
                logger.warning('Invalid version or corrupted offset found. '
                               'Attempting recovery.')
        if message and message.valid:
            recovering = False
            yield message
        elif recovering:
            # bump offset and try again
            offset += 1
            continue
        elif message is not None:
            if len(message) + offset == len(payload):
                # If this is the last message,
                # it's OK to drop it if it's truncated.
                logger.debug('Discarding partial message '
                            '(expected %s bytes, got %s): %s',
                            length, len(message), message)
                return
            else:
                raise AssertionError(
                    "Length of %s (%s) does not match it's "
                    "stated frame size of %s" % (message, len(message), length)
                )
        offset += length

    # Finally, if we can't recover, raise the original exception
    if recovering:
        raise exception


def write_request_header(request, topic, partition):
    request.frame(2, topic)
    request.pack(4, partition)
    return request


class Connection(object):
    """A socket connection to Kafka."""

    def __init__(self, host, port, timeout):
        self.host = host
        self.port = port
        self.timeout = timeout
        self._socket = None

    def __del__(self):
        self.disconnect()

    @property
    def connected(self):
        """
        Do we think the socket is open.
        """
        return self._socket is not None

    def connect(self):
        """
        Connect to the broker.
        """
        self._socket = socket.create_connection(
            (self.host, self.port),
            timeout=self.timeout
        )

    def disconnect(self):
        """
        Disconnect from the Kafka broker.
        """
        try:
            self._socket.close()
        except IOError:
            pass
        finally:
            self._socket = None

    def reconnect(self):
        self.disconnect()
        self.connect()

    def request(self, request):
        """Make a request using the data in `request`.

        :param request: Request data.
        :type request: :class:`samsa.utils.structuredio.StructuredBytesIO`

        """
        # TODO: Retry/reconnect on failure?
        self._socket.sendall(str(request.wrap(4)))

    def response(self, future):
        """Wait for a response and assign to future.

        :param future: Where to assign response data.
        :type future: :class:`samsa.handlers.ResponseFuture`

        """
        try:
            response = recv_framed(self._socket, ResponseFrameHeader)
        except SocketDisconnectedError:
            self.disconnect()
            raise

        header = ResponseErrorHeader.unpack_from(buffer(response))
        if header.error:
            exception_class = ERROR_CODES.get(header.error, -1)
            # TODO: Add better error messaging.
            future.set_error(exception_class)
        else:
            future.set_response(buffer(response, ResponseErrorHeader.size))


class Client(object):
    """
    Low-level Kafka protocol client.

    :param host: broker host
    :param port: broker port number
    :param timeout: socket timeout
    """
    def __init__(self, host, handler, port=9092, timeout=30, autoconnect=True):
        self.connection = Connection(host, port, timeout)
        self.handler = handlers.RequestHandler(handler, self.connection)
        if autoconnect:
            self.connect()

    def connect(self):
        self.connection.connect()
        self.handler.start()

    def disconnect(self):
        self.handler.stop()
        self.connection.disconnect()

    __repr__ = attribute_repr('connection')

    # Protocol Implementation

    def produce(self, topic, partition, messages,
                version=DEFAULT_VERSION, **kwargs):
        """
        Sends messages to the broker on a single topic/partition combination.

        >>> client.produce('topic', 0, ('message',))

        :param topic: topic name
        :param partition: partition ID
        :param messages: the messages to be sent
        :type messages: list, generator, or other iterable of strings
        :param version: version of message encoding
        :type version: int
        :param \*\*kwargs: extra (version-specific) keyword arguments
                           to pass to message encoder
        """
        request = StructuredBytesIO()
        request.pack(2, REQUEST_TYPE_PRODUCE)
        write_request_header(request, topic, partition)
        request.write(Message.encode(messages, version=version, **kwargs))
        return self.handler.request(request, has_response=False)

    def multiproduce(self, data, version=DEFAULT_VERSION, **kwargs):
        """
        Sends messages to the broker on multiple topics and/or partitions.

        >>> client.produce((
        ...    ('topic-1', 0, ('message',)),
        ...    ('topic-2', 0, ('message', 'message',)),
        ... ))

        :param data: sequence of 3-tuples of the format
                     ``(topic, partition, messages)``
        :type data: list, generator, or other iterable
        :param version: version of message encoding
        :type version: int
        :param \*\*kwargs: extra (version-specific) keyword arguments
                           to pass to message encoder
        """
        payloads = []
        for topic, partition, messages in data:
            payload = StructuredBytesIO()
            write_request_header(payload, topic, partition)
            payload.write(Message.encode(messages, version=version, **kwargs))
            payloads.append(payload)

        request = StructuredBytesIO()
        request.pack(2, REQUEST_TYPE_MULTIPRODUCE)
        request.pack(2, len(payloads))
        for payload in payloads:
            request.write(payload)
        return self.handler.request(request, has_response=False)

    def fetch(self, topic, partition, offset, size):
        """
        Fetches messages from the broker on a single topic/partition.

        >>> for offset, message in client.fetch('test', 0, 0, 1000):
        ...     print offset, message
        0L 'hello world'
        20L 'hello world'

        :param topic: topic name
        :param partition: partition ID
        :param offset: offset to begin read
        :type offset: integer
        :param size: the maximum number of bytes to return
        :rtype: generator of 2-tuples in ``(offset, message)`` format
        """
        # TODO: Document failure modes.
        request = StructuredBytesIO()
        request.pack(2, REQUEST_TYPE_FETCH)
        write_request_header(request, topic, partition)
        request.pack(8, offset)
        request.pack(4, size)

        response = self.handler.request(request)

        try:
            return decode_messages(response.get(), from_offset=offset)
        except SocketDisconnectedError:
            return []

    def multifetch(self, data):
        """
        Fetches messages from the broker on multiple topics/partitions.

        >>> topics = (
        ...     ('topic-1', 0, 0, 1000),
        ...     ('topic-2', 0, 0, 1000),
        ... )
        >>> for i, response in enumerate(client.fetch(topics)):
        ...     print 'response:', i
        ...     for offset, message in messages:
        ...         print offset, message
        response 0
        0L 'hello world'
        20L 'hello world'
        response 1
        0L 'hello world'
        20L 'hello world'

        :param data: sequence of 4-tuples of the format
                     ``(topic, partition, offset, size)``
                     For more information, see :meth:`Client.fetch`.
        :rtype: generator of fetch responses (message generators).
            For more information, see :meth:`Client.fetch`.
        """
        payloads = []
        from_offsets = []
        for topic, partition, offset, size in data:
            payload = StructuredBytesIO()
            write_request_header(payload, topic, partition)
            from_offsets.append(offset)
            payload.pack(8, offset)
            payload.pack(4, size)
            payloads.append(payload)

        request = StructuredBytesIO()
        request.pack(2, REQUEST_TYPE_MULTIFETCH)
        request.pack(2, len(payloads))
        for payload in payloads:
            request.write(payload)
        response = self.handler.request(request)
        return decode_message_sets(response.get(), from_offsets)

    def offsets(self, topic, partition, time, max):
        """
        Returns message offsets before a certain time for the given
        topic/partition.

        >>> client.offsets('test', 0, OFFSET_EARLIEST, 1)
        [0]

        :param topic: topic name
        :param partition: partition ID
        :param time: the time in milliseconds since the UNIX epoch, or either
            ``OFFSET_EARLIEST`` or ``OFFSET_LATEST``.
        :type time: integer
        :param max: the maximum number of offsets to return
        :rtype: list of offsets
        """
        request = StructuredBytesIO()
        request.pack(2, REQUEST_TYPE_OFFSETS)
        write_request_header(request, topic, partition)
        request.pack(8, time)
        request.pack(4, max)
        response = self.handler.request(request)
        (count,) = OffsetsResponseHeader.unpack_from(response.get())
        offsets = []
        for i in xrange(0, count):
            offsets.append(Offset.unpack_from(response.get(),
                offset=OffsetsResponseHeader.size + (i * Offset.size)).value)
        return offsets



#!/usr/bin/env python
"""
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys

from setuptools import setup, find_packages

from samsa import __version__

try:
    import multiprocessing
except ImportError:
    pass

install_requires = [
    'kazoo<1.2' # TODO: Remove once they fix the pip upload
]

lint_requires = [
    'pep8',
    'pyflakes'
]

tests_require = ['mock', 'nose', 'unittest2']
dependency_links = []
setup_requires = []
if 'nosetests' in sys.argv[1:]:
    setup_requires.append('nose')

setup(
    name='samsa',
    version=__version__,
    author='Ted Kaemming, Matthew Hooker, and Keith Bourgoin',
    author_email='samsa-users@googlegroups.com',
    url='https://github.com/getsamsa/samsa',
    description='Featureful Kafka client.',
    license='Apache License 2.0',
    packages=find_packages(),
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    extras_require={
        'test': tests_require,
        'all': install_requires + tests_require,
        'docs': ['sphinx'] + tests_require,
        'lint': lint_requires
    },
    dependency_links=dependency_links,
    zip_safe=False,
    test_suite='nose.collector',
    include_package_data=True,
)



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import io
from struct import Struct


class StructuredBytesIO(io.BytesIO):
    BYTES = {
        1: Struct('>b'),
        2: Struct('>h'),
        4: Struct('>i'),
        8: Struct('>q'),
    }

    def __len__(self):
        return len(self.getvalue())

    def __str__(self):
        return self.getvalue()

    def write(self, value, *args, **kwargs):
        if isinstance(value, StructuredBytesIO):
            value = str(value)
        return super(StructuredBytesIO, self).write(value, *args, **kwargs)

    def pack(self, size, *args):
        return self.write(self.BYTES[size].pack(*args))

    def unpack(self, size):
        return self.BYTES[size].unpack(self.read(self.BYTES[size].size))[0]

    def frame(self, size, value):
        value = str(value)
        return self.pack(size, len(value)) + self.write(value)

    def unframe(self, size, validate=True):
        length = self.unpack(size)
        value = self.read(int(length))
        if validate and len(value) != length:
            raise ValueError('Payload length does not match length specified '
                'in header')
        return value

    def wrap(self, size):
        envelope = self.__class__()
        envelope.frame(size, self)
        envelope.seek(0)
        return envelope

    def unwrap(self, size, validate=True):
        payload = self.__class__()
        payload.write(self.unframe(size, validate))
        payload.seek(0)
        return payload

    def print_debug(self):
        import string
        offset = self.tell()
        print ''.join(["%02X " % ord(x) for x in self.read()]).strip()
        self.seek(offset)

        def filt(c):
            if ord(c) < 32 or ord(c) > 126 or c not in string.printable:
                return '?'
            return c
        print '  '.join(map(filt, self.read()))

        self.seek(offset)



from logging import Handler


class KafkaHandler(Handler):
    """
    Basic Kafka-based logging handler.

    :param topic: topic to publish to
    :type topic: :class:`samsa.topics.Topic`
    """
    def __init__(self, topic, *args, **kwargs):
        Handler.__init__(self, *args, **kwargs)  # avoid `super` for 2.6
        self.topic = topic

    def emit(self, record):
        self.topic.publish(self.format(record))



__license__ = """
Copyright 2012 DISQUS

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import time
import unittest2

from samsa.utils.log import get_logger_for_function


class TestCase(unittest2.TestCase):
    """
    A :class:`~unittest2.TestCase` subclass that extra useful test methods.
    """
    def assertPassesWithMultipleAttempts(self, fn, attempts, timeout=1,
            backoff=None, logger=None):
        """
        Executes a callable multiple times, until either the callable exits
        without raising an ``AssertionError``, or the maximum number of
        attempts is exceeded.

        :param fn: test callable
        :type fn: callable
        :param attempts: maximum number of attempts allowed before failing
        :type attempts: ``int``
        :param timeout: number of seconds to wait between attempts
        :type timeout: ``int``
        :param backoff: custom backoff algorithm function that accepts the
            parameters ``(attempt, timeout)``
        :type backoff: callable

        :raises: ``AssertionError`` if an error is encountered on the final
            attempt
        """
        if backoff is None:
            backoff = lambda attempt, timeout: timeout

        if logger is None:
            logger = get_logger_for_function(fn)

        for attempt in xrange(1, attempts + 1):
            logger.debug('Starting attempt %s for %s...', attempt, fn)
            try:
                fn()
                logger.info('Passed attempt %s for %s', attempt, fn)
                break
            except AssertionError:
                if attempt < attempts:
                    wait = backoff(attempt, timeout)
                    logger.exception('Failed attempt %s for %s, waiting for '
                        '%s seconds', attempt, fn, wait)
                    time.sleep(wait)
                else:
                    raise



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import time

from collections import deque
from kazoo.exceptions import NodeExistsException, NoNodeException
from functools import partial
import Queue

from samsa.config import ConsumerConfig
from samsa.exceptions import PartitionOwnedError
from samsa.partitions import Partition


logger = logging.getLogger(__name__)


class OwnedPartition(Partition):
    """Represents a partition as a consumer group sees it.

    Each partition owns a long-lived fetch thread that keeps
    and internal message queue full. It also has a simple
    monitor thread restart the fetch thread if it crashes

    """

    def __init__(self, partition, group):
        """Initialize with a partition and consumer group.

        :param partition: Partition to initialize with.
        :type partition: :class:`samsa.partitions.Partition`.
        :param group: Group that owns this partition.
        :type group: str.

        """
        super(OwnedPartition, self).__init__(
            partition.cluster, partition.topic,
            partition.broker, partition.number
        )

        self.config = ConsumerConfig().build()
        self.group = group
        self.path = "/consumers/%s/offsets/%s/%s-%s" % (
            self.group, self.topic.name,
            self.broker.id, self.number
        )

        # _current_offset is offset after the message most recently consumed
        try:
            offset, stat = self.cluster.zookeeper.get(self.path)
            self._current_offset = int(offset) if offset else 0
        except NoNodeException:
            self._current_offset = 0
            self.cluster.zookeeper.ensure_path(self.path)

        # the next offset to fetch --  initially the same as current
        self._next_offset = self._current_offset
        self._message_queue = Queue.Queue(self.config['queuedchunks_max'])

        # Fetch and monitor threads
        self._monitor_interval = 1 #: Time between monitoring checks
        self._monitor_thread = None #: Monitors the fetch thread
        self._fetch_thread = None #: Keep the partition queue full

        # Flag to stop returning messages or fetching more
        self._running = False


    @property
    def offset(self):
        return self._current_offset

    def _fetch(self):
        """Entry point for the fetch thread

        Runs as a long-lived thread that will keep the internal queue full.

        """
        while True:
            if not self._running:
                return

            # Fetch more messages
            messages = super(OwnedPartition, self).fetch(
                self._next_offset,
                self.config['fetch_size']
            )

            # Iterate over the list and try to put into _message_queue
            # but be sure to check _running even when blocking
            messages = deque(messages)
            while len(messages) > 0:
                message = messages.popleft() # deque adds items left-to-right
                if not self._running:
                    return # thread death
                try:
                    self._message_queue.put(
                        message,
                        block=True,
                        timeout=self._monitor_interval
                    )
                    self._next_offset = message.next_offset
                except Queue.Full:
                    # Put it back, check _running and then try again
                    messages.appendleft(message)

    def _monitor(self):
        """Makes sure that the fetch thread restarts on crash

        """
        while True:
            if not self._running:
                return
            if not self._fetch_thread or not self._fetch_thread.is_alive():
                self._fetch_thread = self.cluster.handler.spawn(target=self._fetch)
            time.sleep(self._monitor_interval)

    def empty(self):
        """True if there are no messages.

        """
        return self._message_queue.empty()

    def next_message(self, block=True, timeout=None):
        """Retrieve the next message for this partition.

        Returns None if the partition is stopped, or there are no messages
        within the timeout (or when not blocking)

        :param block: whether to block at all
        :param timeout: block for timeout if integer, or indefinitely if None.

        """
        if not self._running:
            return None

        try:
            message = self._message_queue.get(block=block, timeout=timeout)
            self._current_offset = message.next_offset
            return message.payload
        except Queue.Empty:
            return None

    def commit_offset(self):
        """Commit current offset to zookeeper.

        """
        self.cluster.zookeeper.set(self.path, str(self._current_offset))

    def start(self):
        """Start the fetch thread

        """
        self._running = True
        self._monitor_thread = self.cluster.handler.spawn(target=self._monitor)

    def stop(self):
        """Stop the fetch thread.

        """
        self._running = False

    def __del__(self):
        self.stop()


class PartitionOwnerRegistry(object):
    """Manages the Partition Owner Registry for a particular Consumer.

    This also helps manage access to the owned partitions. The Registry
    provides the ``get_ready_partition`` method that will give a consumer
    the next partition that has messages ready to be read. Using this,
    a consumer can iterate over active partitions without ever having to
    block while waiting on a less active one.

    Using the ready partition functionality, it's possible to multiplex
    between available partitions in a lock-free, highly available manner.
    To improve throughput, ideally a client would read several message from
    any given partition before moving on to the next.

    """

    def __init__(self, consumer, cluster, topic, group):
        """
        :param consumer: consumer which owns these partitions.
        :type consumer: :class:`samsa.consumer.Consumer`.
        :param cluster:
        :type cluster: :class:`samsa.cluster.Cluster`.
        :param topic: The topic the partitions belong to.
        :type topic: :class:`samsa.topics.Topic`.
        :param group: The group the partitions belong to.
        :type group: str.

        """
        self.consumer = consumer
        self.cluster = cluster
        self.topic = topic

        self.Partition = partial(OwnedPartition, group=group)
        self.path = '/consumers/%s/owners/%s' % (group, topic.name)
        self.cluster.zookeeper.ensure_path(self.path)
        self._partitions = set([])
        self._ready_queue = Queue.Queue()
        self._watch_interval = 0.5

    def get(self):
        """Get all owned partitions.

        """
        return self._partitions

    def remove(self, partitions):
        """Remove `partitions` from the registry.

        :param partitions: partitions to remove.
        :type partitions: iterable of :class:`samsa.partitions.Partition`.

        """
        for p in partitions:
            assert p in self._partitions
            p.stop() # TODO: Fix stop so it happens here instead of in consumer
            self.cluster.zookeeper.delete(self._path_from_partition(p))
            self._partitions.remove(p)

    def get_ready_partition(self, block=True, timeout=None):
        """Get the next partition with messages to read

        Returns None if none are available and nonblocking or timeout expires

        """
        # N.B.: Immediately setting the dequeued event means the watch
        # thread will unblock, see there are pending messages, and requeue
        # the partition. If a partition is exhausted during its read, then
        # it'll be empty the next time it's dequeued.
        #
        # This is by design. By not waiting to requeue the partition
        # there's no need to keep track of when it starts and stops usage.
        # That cuts down on monitoring overhead, reduces potential bugs due
        # to not letting go of the partition, and most importantly means
        # we can do the entire thing lock-free, using only Events.
        #
        # Also note the immediate-requeue only happens once. After that,
        # the watch thread will block until more messages are fetched.
        while True:
            try:
                partition, dequeued  = self._ready_queue.get(
                    block=block, timeout=timeout
                )
                dequeued.set()
            except Queue.Empty:
                return None

            # only return active partitions that still have messages
            if partition not in self._partitions or partition.empty():
                continue

            return partition

    def add(self, partitions):
        """Add `partitions` to the registry.

        :param partitions: partitions to add.
        :type partitions: iterable of :class:`samsa.partitions.Partition`.

        """
        for p in partitions:
            try:
                self.cluster.zookeeper.create(
                    self._path_from_partition(p), self.consumer.id,
                    ephemeral=True
                )
            except NodeExistsException:
                raise PartitionOwnedError(p)

            # Add to self.partition and start the monitor
            partition = self.Partition(p)
            partition.start()
            self._partitions.add(partition)
            self.cluster.handler.spawn(
                target=self._watch_partition, args=[partition]
            )

    def _path_from_partition(self, p):
        return "%s/%s-%s" % (self.path, p.broker.id, p.number)

    def _watch_partition(self, partition):
        """Watch the partition and put it in the ready_queue when it has data

        This watch thread (and ready queue) are how we can implement the
        partition multiplexing without lock-free and with very fast data
        availability

        """
        # N.B.: Everything in the main loop needs a timeout so it
        # can check for partition membership. Otherwise it never dies.
        import threading # TODO: Obv incompatible with gevent.
        dequeued = threading.Event()
        dequeued.set()
        while True:
            if partition not in self._partitions:
                return

            # Wait for it to be removed from the queue
            dequeued.wait(self._watch_interval)
            if not dequeued.is_set():
                continue

            # Wait for messages. Use of not_empty is from Queue.get source
            # TODO: Also very incompatible with gevent
            queue = partition._message_queue
            queue.not_empty.acquire()
            try:
                if queue._qsize() == 0:
                    queue.not_empty.wait(self._watch_interval)
                if queue._qsize() == 0:
                    continue # check partition membership
            finally:
                queue.not_empty.release()

            # ``.clear`` before ``.put`` avoids a race condition
            dequeued.clear()
            self._ready_queue.put((partition, dequeued))



__license__ = """
Copyright 2012 DISQUS
Copyright 2013 Parse.ly, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import functools
import logging


logger = logging.getLogger(__name__)


def requires_configuration(method):
    """
    A method decorator for objects that derive from
    :class:`DelayedConfiguration` that ensures methods that require
    configuration have the appropriate state before being invoked.
    """
    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        if not self._configured:
            logger.debug('%s requires configuration before %s may be invoked. '
                         'Configuring now.',
                         self, method)
            self._configured = True
            self._configure()
        return method(self, *args, **kwargs)
    return wrapped


class DelayedConfiguration(object):
    """
    A mixin class for objects that can be instantiated without their full
    configuration available.

    Subclasses must implement :meth:`_configure`, which will be called once
    to bootstrap the configuration for the each instance.
    """
    _configured = False

    def _configure(self):
        raise NotImplementedError
